package io.github.taha_cmd.api.http;

import com.squareup.moshi.JsonAdapter;
import com.squareup.moshi.Moshi;
import io.github.taha_cmd.SignPathClientSettings;
import io.github.taha_cmd.SignPathClientException;
import io.github.taha_cmd.SignPathClientSimpleLogger;
import io.github.taha_cmd.api.model.SigningRequest;
import io.github.taha_cmd.api.model.SigningRequestSubmitResponse;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Map;

import okhttp3.MediaType;
import okhttp3.MultipartBody;
import okhttp3.OkHttpClient;
import okhttp3.Request;
import okhttp3.RequestBody;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;

/**
 *
 * @author rober
 */
public class SignPathApiHttpClient {
    public static final String USER_AGENT_HEADER_NAME = "User-Agent";
    public static final String SIGNPATH_ORIGIN_HEADER_NAME = "SignPath-Origin";
    public static final String SIGNPATH_ALLOW_IGNORE_ORIGIN_HEADER_VALUE = "allow-ignore";
    private static final String ARTIFACT_PARAM_NAME = "Artifact";

    private final OkHttpClient okHttpClient;
    private final String signPathApiBaseUrl;
    private final SignPathClientSettings signPathClientSettings;
    private final SignPathClientSimpleLogger logger;

    public SignPathApiHttpClient(String signPathApiBaseUrl, SignPathClientSimpleLogger logger, SignPathClientSettings signPathClientSettings) {
        this.signPathApiBaseUrl = signPathApiBaseUrl;
        this.logger = logger;
        this.signPathClientSettings = signPathClientSettings;
        
        logger.log("User agent: " + signPathClientSettings.getUserAgent());
        
        int requestTimeoutInSeconds = signPathClientSettings.getUploadAndDownloadRequestTimeoutInSeconds();
        this.okHttpClient = HttpClientFactory.createHttpClient(4, requestTimeoutInSeconds, logger, this::formatResponseError);
    }

    public SigningRequest getSigningRequestHttp(String apiToken,
            String organizationId,
            String signingRequestId) {

        String url = "v1/{OrganizationId}/SigningRequests/{SigningRequestId}";
        url = url.replace("{OrganizationId}", organizationId);
        url = url.replace("{SigningRequestId}", signingRequestId);

        Request request = new Request.Builder()
                .url(this.signPathApiBaseUrl + url)
                .addHeader("Authorization", this.buildAuthorizationHeaderValue(apiToken, null))
                .addHeader(USER_AGENT_HEADER_NAME, this.signPathClientSettings.getUserAgent())
                .build();

        okhttp3.Call call = okHttpClient.newCall(request);
        okhttp3.Response response = executeOkHttpCall(call);
        SigningRequest signingRequest = new SigningRequest();
        if (response.isSuccessful() && response.body() != null) {
            try {
                Moshi moshi = new Moshi.Builder().build();
                JsonAdapter<SigningRequest> jsonAdapter = moshi.adapter(SigningRequest.class).lenient();

                signingRequest = jsonAdapter.fromJson(response.body().string());
            } catch (IOException ex) {
                logger.log(ex.getMessage(), ex);
            }
        }
        return signingRequest;
    }

    public String submitSigningRequestHttp(
            String apiToken,
            String tbsToken,
            String organizationId,
            File artifact,
            String projectSlug,
            String signingPolicySlug,
            String artifactConfigurationSlug,
            String description,
            boolean allowIgnoreOrigin,
            Map<String, String> origin,
            Map<String, String> userDefinedParameters) {

        String url = "v1/{OrganizationId}/SigningRequests";
        url = url.replace("{OrganizationId}", organizationId);

        MultipartBody.Builder requestBuilder = new MultipartBody.Builder()
                .setType(MultipartBody.FORM)
                .addFormDataPart(ARTIFACT_PARAM_NAME, artifact.getName(),
                        RequestBody.create(MediaType.parse("application/octet-stream"), artifact))
                .addFormDataPart("ProjectSlug", projectSlug)
                .addFormDataPart("SigningPolicySlug", signingPolicySlug);

        if (artifactConfigurationSlug != null && !artifactConfigurationSlug.isEmpty()) {
            requestBuilder.addFormDataPart("ArtifactConfigurationSlug", artifactConfigurationSlug);
        }
        if (description != null && !description.isEmpty()) {
            requestBuilder.addFormDataPart("Description", description);
        }

        addMapToRequestBody("Origin", origin, requestBuilder);
        addMapToRequestBody("Parameters", userDefinedParameters, requestBuilder);

        RequestBody requestBody = requestBuilder.build();

        Request.Builder builder = new Request.Builder()
                .url(this.signPathApiBaseUrl + url)
                .addHeader("Authorization", this.buildAuthorizationHeaderValue(apiToken, tbsToken))
                .addHeader(USER_AGENT_HEADER_NAME, this.signPathClientSettings.getUserAgent())
                .post(requestBody);

        if (allowIgnoreOrigin) {
            builder.addHeader(SIGNPATH_ORIGIN_HEADER_NAME, SIGNPATH_ALLOW_IGNORE_ORIGIN_HEADER_VALUE);
        }

        Request request = builder.build();

        okhttp3.Call call = okHttpClient.newCall(request);
        okhttp3.Response response = executeOkHttpCall(call);
        SigningRequestSubmitResponse signingRequestSubmitResponse = new SigningRequestSubmitResponse();
        if (response.isSuccessful() && response.body() != null) {
            try {
                Moshi moshi = new Moshi.Builder().build();
                JsonAdapter<SigningRequestSubmitResponse> jsonAdapter = moshi.adapter(SigningRequestSubmitResponse.class).lenient();
                signingRequestSubmitResponse = jsonAdapter.fromJson(response.body().string());
            } catch (IOException ex) {
                logger.log(ex.getMessage(), ex);
            }
        }
        return signingRequestSubmitResponse.getSigningRequestId();
    }

    public void downloadSignedArtifactHttp(
            String apiToken,
            String organizationId,
            String signingRequestId,
            File artifactTargetFile) {

        String url = "v1/{OrganizationId}/SigningRequests/{SigningRequestId}/SignedArtifact";
        url = url.replace("{OrganizationId}", organizationId);
        url = url.replace("{SigningRequestId}", signingRequestId);

        Request request = new Request.Builder()
                .url(this.signPathApiBaseUrl + url)
                .addHeader("Authorization", this.buildAuthorizationHeaderValue(apiToken, null))
                .addHeader(USER_AGENT_HEADER_NAME, this.signPathClientSettings.getUserAgent())
                .build();

        okhttp3.Call call = okHttpClient.newCall(request);
        okhttp3.Response response = executeOkHttpCall(call);

        try ( InputStream stream = response.body().byteStream();  FileOutputStream outStream = new FileOutputStream(artifactTargetFile);) {
            copyStream(stream, outStream);
            outStream.close();
        } catch (IOException ex) {
            logger.log(ex.getMessage(), ex);
            throw new SignPathClientException("Artifact download error", ex);
        }
    }

    private void addMapToRequestBody(String prefix, Map<String, String> map, MultipartBody.Builder requestBuilder) {
        if (map != null) {
            prefix = prefix + ".";
            for (String key : map.keySet()) {
                String value = map.get(key);
                if (value.startsWith("@")) {
                    // value is a file path
                    // in our case it only could be an artifact body path
                    String filePath = value.substring(1);
                    requestBuilder.addFormDataPart(prefix + key, ARTIFACT_PARAM_NAME,
                            RequestBody.create(MediaType.parse("application/octet-stream"), new File(filePath)));

                } else {
                    requestBuilder.addFormDataPart(prefix + key, value);
                }
            }
        }
    }

    private okhttp3.Response executeOkHttpCall(okhttp3.Call call) throws SignPathClientException {
            try {
                return call.execute();
            } catch (IOException ex) {
                throw new SignPathClientException("SignPath API communication error.", ex);
            }
    }

    private String formatResponseError (okhttp3.Response response) {
        logger.log("No success response.");
        String responseBody = "";
        
        try {
          responseBody = response.body().string();
        } catch (IOException ex) {
          logger.log(ex.getMessage(), ex);
        }

        String additionalReason = "";
        if (401 == response.code()) {
          additionalReason = " Did you provide the correct ApiToken?";
        }
        else if(403 == response.code() && !responseBody.contains("feature is disabled")) {
          additionalReason = " Did you add the CI user to the list of submitters in the specified signing policy? Did you provide the correct OrganizationId? In case you are using a trusted build system, did you link it to the specified project?";
        }

        String serverMessage = "";
        if (!StringUtils.isEmpty(responseBody)) {
          serverMessage = " (Server reported the following message: '" + responseBody + "')";
        }

        return String.format("Error %1$s.%2$s%3$s", response.code(),additionalReason, serverMessage);
    }

    private String buildAuthorizationHeaderValue(String apiToken, String tbsToken) {
        Validate.notEmpty(apiToken);

        if (tbsToken == null || tbsToken.isEmpty()) {
            return String.format("Bearer %s", apiToken);
        }
        return String.format("Bearer %s:%s", apiToken, tbsToken);
    }

    private void copyStream(InputStream input, OutputStream output) throws IOException {
        byte[] buf = new byte[8192];
        int length;
        while ((length = input.read(buf)) > 0) {
            output.write(buf, 0, length);
        }
    }
}
